from dataclasses import dataclass
from typing import Optional, Union

from arclet.alconna import Alconna, Arparma
from arclet.letoderea import Contexts, Provider, make_event

from ..message import MessageChain
from ..session import Session
from .base import Reply


@dataclass
@make_event(name="entari.event/command/execute")
class CommandExecute:
    command: Union[str, MessageChain]

    async def gather(self, context: Contexts):
        if isinstance(self.command, str):
            context["command"] = MessageChain(self.command)
        else:
            context["command"] = self.command

    class CommandProvider(Provider[MessageChain]):
        async def __call__(self, context: Contexts):
            return context.get("command")

    __result_type__: "type[str | MessageChain]" = Union[str, MessageChain]


@dataclass
@make_event(name="entari.event/command/output")
class CommandOutput:
    """依据输出信息的类型，将字符串转换为消息对象以便发送。"""

    session: Session
    command: Alconna
    type: str
    content: str

    __result_type__: "type[bool | str | MessageChain]" = Union[bool, str, MessageChain]


@dataclass
@make_event(name="entari.event/command/before_parse")
class CommandReceive:
    """尝试解析命令时调用"""

    session: Session
    command: Alconna
    content: MessageChain
    reply: Optional[Reply] = None

    __result_type__: "type[MessageChain]" = MessageChain


@dataclass
@make_event(name="entari.event/command/after_parse")
class CommandParse:
    """解析完命令后调用"""

    session: Session
    command: Alconna
    result: Arparma

    __result_type__: "type[Arparma | bool]" = Union[Arparma, bool]



from __future__ import annotations

import inspect
import logging
import sys
from typing import TYPE_CHECKING

from loguru import logger

if TYPE_CHECKING:
    from loguru import Logger, Record


class LoggerManager:
    def __init__(self):
        self.loggers: dict[str, Logger] = {}
        self.fork("[core]")
        self.fork("[plugin]")
        self.fork("[message]")

    def fork(self, child_name: str):
        patched = logger.patch(lambda r: r.update(name=child_name))
        patched = patched.bind(name=child_name)
        self.loggers[child_name] = patched
        return patched

    @property
    def core(self):
        return self.loggers["[core]"].opt(colors=True)

    @property
    def plugin(self):
        return self.loggers["[plugin]"].opt(colors=True)

    @property
    def message(self):
        return self.loggers["[message]"]

    def wrapper(self, name: str, color: str = "blue"):
        patched = logger.patch(
            lambda r: r.update(
                name="entari", extra=r["extra"] | {"entari_plugin_name": name, "entari_plugin_color": color}
            )
        )
        patched = patched.bind(name=f"plugins.{name}")
        self.loggers[f"plugin.{name}"] = patched
        return patched

    @staticmethod
    def set_level(level: str | int):
        if isinstance(level, str):
            level = level.upper()
        logging.basicConfig(
            handlers=[LoguruHandler()],
            level=level,
            format="%(asctime)s | %(name)s[%(levelname)s]: %(message)s",
        )
        logger.configure(extra={"entari_log_level": level}, patcher=_hidden_upsteam)


class LoguruHandler(logging.Handler):  # pragma: no cover
    """logging 与 loguru 之间的桥梁，将 logging 的日志转发到 loguru。"""

    def emit(self, record: logging.LogRecord):
        try:
            level = logger.level(record.levelname).name
        except ValueError:
            level = record.levelno

        frame, depth = inspect.currentframe(), 0
        while frame and (depth == 0 or frame.f_code.co_filename == logging.__file__):
            frame = frame.f_back
            depth += 1

        logger.opt(depth=depth, exception=record.exc_info).log(level, record.getMessage())


logging.basicConfig(
    handlers=[LoguruHandler()],
    level="INFO",
    format="%(asctime)s | %(name)s[%(levelname)s]: %(message)s",
)


def default_filter(record):
    if record["name"].startswith("launart"):
        return record["level"].no >= logger.level("SUCCESS").no
    log_level = record["extra"].get("entari_log_level", "INFO")
    levelno = logger.level(log_level).no if isinstance(log_level, str) else log_level
    return record["level"].no >= levelno


def _custom_format(record: Record):
    if "entari_plugin_name" in record["extra"]:
        plugin = (
            f" <{record['extra']['entari_plugin_color']}>"
            f"{record['extra']['entari_plugin_name']}"
            f"</{record['extra']['entari_plugin_color']}>"
        )
    else:
        plugin = ""
    return (
        f"<lk>{{time:YYYY-MM-DD HH:mm:ss}}</lk> <lvl>{{level}}</lvl> | <m><u>{{name}}</u></m>"
        f"{plugin} <lvl>{{message}}</lvl>\n"
    )


logger.remove()
logger_id = logger.add(
    sys.stdout,
    level=0,
    diagnose=True,
    backtrace=True,
    colorize=True,
    filter=default_filter,
    format=_custom_format,
)
"""默认日志处理器 id"""


def _hidden_upsteam(record: Record):
    if record["name"].startswith("satori"):  # type: ignore
        record["name"] = "satori"
    if record["name"].startswith("launart"):  # type: ignore
        record["name"] = "launart"


logger.configure(patcher=_hidden_upsteam)
log = LoggerManager()


__all__ = ["log"]



import asyncio
from pathlib import Path
from typing import Union

from arclet.letoderea import es
from launart import Launart, Service, any_completed
from launart.status import Phase

try:
    from watchfiles import PythonFilter, awatch
except ModuleNotFoundError:
    raise ImportError("Please install `watchfiles` first. Install with `pip install arclet-entari[reload]`")

from arclet.entari import add_service, declare_static, load_plugin, metadata, plugin_config, unload_plugin
from arclet.entari.config import BasicConfModel, EntariConfig, field
from arclet.entari.event.config import ConfigReload
from arclet.entari.logger import log
from arclet.entari.plugin import find_plugin, find_plugin_by_file

declare_static()


class Config(BasicConfModel):
    watch_dirs: list[Union[str, Path]] = field(default_factory=lambda: ["."])
    watch_config: bool = False


metadata(
    "AutoReload",
    author=["RF-Tar-Railt <rf_tar_railt@qq.com>"],
    description="Auto reload plugins when files changed",
    config=Config,
)


logger = log.wrapper("[AutoReload]").opt(colors=True)


def detect_filter_change(old: dict, new: dict):
    added = set(new) - set(old)
    removed = set(old) - set(new)
    changed = {key for key in set(new) & set(old) if new[key] != old[key]}
    return "allow" in (added | removed | changed) or "$deny" in (added | removed | changed)


class Watcher(Service):
    id = "watcher"

    @property
    def required(self) -> set[str]:
        return set()

    @property
    def stages(self) -> set[Phase]:
        return {"blocking"}

    def __init__(self, dirs: list[Union[str, Path]], is_watch_config: bool):
        self.dirs = dirs
        self.is_watch_config = is_watch_config
        self.fail = {}
        super().__init__()

    async def watch(self):
        async for event in awatch(*self.dirs, watch_filter=PythonFilter()):
            for change in event:
                if plugin := find_plugin_by_file(change[1]):
                    if plugin.is_static:
                        logger.info(f"Plugin <y>{plugin.id!r}</y> is static, ignored.")
                        continue
                    logger.info(f"Detected change in <blue>{plugin.id!r}</blue>, reloading...")
                    pid = plugin.id
                    del plugin
                    unload_plugin(pid)
                    if plugin := load_plugin(pid):
                        logger.info(f"Reloaded <blue>{plugin.id!r}</blue>")
                        del plugin
                    else:
                        logger.error(f"Failed to reload <blue>{pid!r}</blue>")
                        self.fail[change[1]] = pid
                elif change[1] in self.fail:
                    logger.info(f"Detected change in {change[1]!r} which failed to reload, retrying...")
                    if plugin := load_plugin(self.fail[change[1]]):
                        logger.info(f"Reloaded <blue>{plugin.id!r}</blue>")
                        del plugin
                        del self.fail[change[1]]
                    else:
                        logger.error(f"Failed to reload <blue>{self.fail[change[1]]!r}</blue>")

    async def watch_config(self):
        file = EntariConfig.instance.path.resolve()
        extra = [file.parent.joinpath(path) for path in EntariConfig.instance.plugin_extra_files]
        async for event in awatch(file.absolute(), *(dir_.absolute() for dir_ in extra), recursive=False):
            for change in event:
                if not self.is_watch_config:
                    continue
                if not (
                    (change[0].name == "modified" and Path(change[1]).resolve() == file)
                    or Path(change[1]).resolve() in extra
                    or Path(change[1]).resolve().parent in extra
                ):
                    continue
                logger.info(f"Detected change in {change[1]!r}, reloading config...")

                old_basic = EntariConfig.instance.basic.copy()
                old_plugin = EntariConfig.instance.plugin.copy()
                EntariConfig.instance.reload()
                for key in old_basic:
                    if key in EntariConfig.instance.basic and old_basic[key] != EntariConfig.instance.basic[key]:
                        logger.debug(
                            f"Basic config <y>{key!r}</y> changed from <r>{old_basic[key]!r}</r> "
                            f"to <g>{EntariConfig.instance.basic[key]!r}</g>",
                        )
                        await es.publish(ConfigReload("basic", key, EntariConfig.instance.basic[key], old_basic[key]))
                for key in set(EntariConfig.instance.basic) - set(old_basic):
                    logger.debug(f"Basic config <y>{key!r}</y> appended")
                    await es.publish(ConfigReload("basic", key, EntariConfig.instance.basic[key]))
                for plugin_name in old_plugin:
                    if plugin_name.startswith("$") or plugin_name.startswith("~"):
                        continue
                    pid = plugin_name.replace("::", "arclet.entari.builtins.")
                    if plugin_name not in EntariConfig.instance.plugin:
                        if plugin := find_plugin(pid):
                            del plugin
                            unload_plugin(pid)
                            logger.info(f"Disposed plugin <blue>{pid!r}</blue>")
                        continue
                    if old_plugin[plugin_name] != EntariConfig.instance.plugin[plugin_name]:
                        logger.debug(
                            f"Plugin <y>{plugin_name!r}</y> config changed from <r>{old_plugin[plugin_name]!r}</r> "
                            f"to <g>{EntariConfig.instance.plugin[plugin_name]!r}</g>",
                        )
                        old_conf = old_plugin[plugin_name]
                        new_conf = EntariConfig.instance.plugin[plugin_name]
                        if plugin := find_plugin(pid):
                            filter_changed = detect_filter_change(old_conf, new_conf)
                            if not filter_changed:
                                res = await es.post(
                                    ConfigReload("plugin", plugin_name, new_conf, old_conf),
                                )
                                if res and res.value:
                                    logger.debug(f"Plugin <y>{pid!r}</y> config change handled by itself.")
                                    continue
                            logger.info(f"Detected config of <blue>{pid!r}</blue> changed, reloading...")
                            plugin_file = str(plugin.module.__file__)
                            unload_plugin(pid)
                            if plugin := load_plugin(plugin_name, new_conf):
                                logger.info(f"Reloaded <blue>{plugin.id!r}</blue>")
                                del plugin
                            else:
                                logger.error(f"Failed to reload <blue>{plugin_name!r}</blue>")
                                self.fail[plugin_file] = pid
                        else:
                            logger.info(f"Detected <blue>{pid!r}</blue> appended, loading...")
                            load_plugin(plugin_name, new_conf)
                if new := (set(EntariConfig.instance.plugin) - set(old_plugin)):
                    for plugin_name in new:
                        if plugin_name.startswith("$") or plugin_name.startswith("~"):
                            continue
                        if not (plugin := load_plugin(plugin_name)):
                            continue
                        del plugin

    async def launch(self, manager: Launart):
        async with self.stage("blocking"):
            watch_task = asyncio.create_task(self.watch())
            sigexit_task = asyncio.create_task(manager.status.wait_for_sigexit())
            watch_config_task = asyncio.create_task(self.watch_config())
            done, pending = await any_completed(
                sigexit_task,
                watch_task,
                watch_config_task,
            )
            if sigexit_task in done:
                watch_task.cancel()
                watch_config_task.cancel()
        self.fail.clear()


conf = plugin_config(Config)

add_service(serv := Watcher(conf.watch_dirs, conf.watch_config))


@es.on(ConfigReload)
def handle_config_reload(event: ConfigReload):
    if event.scope != "plugin":
        return
    if event.key not in ("::auto_reload", "arclet.entari.builtins.auto_reload"):
        return
    new_conf = event.plugin_config(Config)
    serv.dirs = new_conf.watch_dirs
    serv.is_watch_config = new_conf.watch_config
    return True



from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Callable, Optional
from typing_extensions import ParamSpec
from weakref import finalize

from nonestorage import user_cache_dir, user_data_dir

from ..event.config import ConfigReload
from ..plugin.model import RootlessPlugin

P = ParamSpec("P")


def _ensure_dir(path: Path) -> None:
    if not path.exists():
        path.mkdir(parents=True, exist_ok=True)
    elif not path.is_dir():
        raise RuntimeError(f"{path} is not a directory")


def _auto_create_dir(func: Callable[P, Path]) -> Callable[P, Path]:
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> Path:
        path = func(*args, **kwargs)
        _ensure_dir(path)
        return path

    return wrapper


class LocalData:
    def __init__(self):
        self.root: Optional[Path] = None
        self.app_name = "Entari"
        self._temp_dir = TemporaryDirectory()
        finalize(self, lambda obj: obj._temp_dir.cleanup(), self)

    def _get_base_cache_dir(self) -> Path:
        return user_cache_dir(self.app_name).resolve() if self.root is None else self.root

    def _get_base_data_dir(self) -> Path:
        return user_data_dir(self.app_name).resolve() if self.root is None else self.root

    @_auto_create_dir
    def get_cache_dir(self, name: Optional[str]) -> Path:
        dir_ = self._get_base_cache_dir()
        return (dir_ / name) if name else dir_

    def get_cache_file(self, name: Optional[str], filename: str) -> Path:
        return self.get_cache_dir(name) / filename

    @_auto_create_dir
    def get_data_dir(self, plugin_name: Optional[str]) -> Path:
        dir_ = self._get_base_data_dir()
        return (dir_ / plugin_name) if plugin_name else dir_

    def get_data_file(self, plugin_name: Optional[str], filename: str) -> Path:
        return self.get_data_dir(plugin_name) / filename

    @_auto_create_dir
    def get_temp_dir(self) -> Path:
        return Path(self._temp_dir.name) / self.app_name

    def get_temp_file(self, filename: str) -> Path:
        return self.get_temp_dir() / filename


local_data = LocalData()


@RootlessPlugin.apply("localdata")
def localdata_apply(plg: RootlessPlugin):
    conf = plg.config
    if "root" in conf:
        local_data.root = Path(conf["root"])
    if "app_name" in conf:
        local_data.app_name = conf["app_name"]
        local_data._temp_dir.cleanup()
        local_data._temp_dir = TemporaryDirectory()

    @plg.use(ConfigReload)
    def reload_config(event: ConfigReload):
        if event.scope != "plugin":
            return
        if event.key != ".localdata":
            return
        conf = event.value
        if "root" in conf:
            local_data.root = Path(conf["root"])
        if "app_name" in conf:
            local_data.app_name = conf["app_name"]
            local_data._temp_dir.cleanup()
            local_data._temp_dir = TemporaryDirectory()



from arclet.alconna import Alconna, Args, CommandMeta
from satori.element import At, Author, Sharp, select

from arclet.entari import MessageEvent, Session, command, metadata
from arclet.entari.command.model import Match

from .i18n import Lang

metadata(
    "inspect",
    ["RF-Tar-Railt <rf_tar_railt@qq.com>"],
    description="Inspect on any user, group or channel",
)


inspect_cmd = Alconna(
    "inspect",
    Args["target?", [At, Sharp]],
    meta=CommandMeta(Lang.entari_plugin_inspect.description(), example="inspect @user\ninspect #channel\ninspect"),
)


SceneNames = {
    "DIRECT": Lang.entari_plugin_inspect.scene.direct,
    "TEXT": Lang.entari_plugin_inspect.scene.text,
    "VOICE": Lang.entari_plugin_inspect.scene.voice,
    "CATEGORY": Lang.entari_plugin_inspect.scene.category,
}


@command.on(inspect_cmd)
async def inspect(session: Session[MessageEvent], target: Match["At | Sharp"]):
    event = session.event
    texts = [
        Lang.entari_plugin_inspect.platform(platform=session.account.platform),
        Lang.entari_plugin_inspect.self(self_id=session.account.self_id),
        Lang.entari_plugin_inspect.scene.name(scene=SceneNames[session.channel.type.name]()),
        Lang.entari_plugin_inspect.guild(guild_id=event.guild.id if event.guild else None),
        Lang.entari_plugin_inspect.channel(channel_id=session.channel.id),
    ]
    if event.quote and (authors := select(event.quote, Author)):
        texts.append(Lang.entari_plugin_inspect.user(user_id=authors[0].id))
        await session.send_message("\n".join(texts))
        return
    if target.available:
        if isinstance(target.result, At) and target.result.id:
            await session.send_message(Lang.entari_plugin_inspect.user(user_id=target.result.id))
        elif isinstance(target.result, Sharp) and target.result.id:
            await session.send_message(Lang.entari_plugin_inspect.channel(channel_id=target.result.id))
        else:
            await session.send_message(Lang.entari_plugin_inspect.invalid())
        return
    texts.append(Lang.entari_plugin_inspect.user(user_id=session.user.id))
    await session.send_message("\n".join(texts))
    return



from __future__ import annotations

from collections.abc import Iterable
from typing import Generic, NoReturn, TypeVar, cast

from arclet.letoderea import HandlerStop, StepOut, es
from satori.client.account import Account
from satori.client.protocol import ApiProtocol
from satori.const import Api
from satori.element import Element
from satori.model import Channel, Guild, Member, MessageReceipt, PageResult, Role, User

from .event.base import FriendRequestEvent, GuildMemberRequestEvent, GuildRequestEvent, MessageEvent, Reply, SatoriEvent
from .event.send import SendRequest, SendResponse
from .message import MessageChain

TEvent = TypeVar("TEvent", bound=SatoriEvent)


class EntariProtocol(ApiProtocol):

    async def send_message(
        self, channel: str | Channel, message: str | Iterable[str | Element], source: SatoriEvent | None = None
    ) -> list[MessageReceipt]:
        """发送消息。返回一个 `MessageReceipt` 对象构成的数组。

        Args:
            channel (str | Channel): 要发送的频道 ID
            message (str | Iterable[str | Element]): 要发送的消息
            source (SatoriEvent | None): 源事件

        Returns:
            list[MessageReceipt]: `MessageReceipt` 对象构成的数组
        """
        channel_id = channel.id if isinstance(channel, Channel) else channel
        return await self.message_create(channel_id=channel_id, content=message, source=source)

    async def send_private_message(
        self, user: str | User, message: str | Iterable[str | Element], source: SatoriEvent | None = None
    ) -> list[MessageReceipt]:
        """发送私聊消息。返回一个 `MessageReceipt` 对象构成的数组。

        Args:
            user (str | User): 要发送的用户 ID
            message (str | Iterable[str | Element]): 要发送的消息
            source (SatoriEvent | None): 源事件

        Returns:
            list[MessageReceipt]: `MessageReceipt` 对象构成的数组
        """
        user_id = user.id if isinstance(user, User) else user
        channel = await self.user_channel_create(user_id=user_id)
        return await self.message_create(channel_id=channel.id, content=message, source=source)

    async def message_create(
        self, channel_id: str, content: str | Iterable[str | Element], source: SatoriEvent | None = None
    ) -> list[MessageReceipt]:
        """发送消息。返回一个 `MessageReceipt` 对象构成的数组。

        Args:
            channel_id (str): 频道 ID
            content (str | Iterable[str | Element]): 消息内容
            source (SatoriEvent | None): 源事件

        Returns:
            list[MessageReceipt]: `MessageReceipt` 对象构成的数组
        """
        msg = MessageChain(content)
        sess = None
        if source:
            sess = Session(self.account, source)
            sess.elements = msg
        if res := await es.post(SendRequest(self.account, channel_id, msg, sess)):
            if not res.value:
                return []
            elif res.value is True:
                pass
            else:
                msg = res.value
        send = str(msg)
        res = await self.call_api(
            Api.MESSAGE_CREATE,
            {"channel_id": channel_id, "content": send},
        )
        res = cast("list[dict]", res)
        resp = [MessageReceipt.parse(i) for i in res]
        await es.publish(SendResponse(self.account, channel_id, msg, resp, sess), SendResponse.__publisher__)
        return resp


class Session(Generic[TEvent]):
    """在 Satori-Python 的 Session 的基础上存储了当次事件的信息"""

    def __init__(self, account: Account[EntariProtocol], event: TEvent):
        self.account = account
        self.event = event
        self.type = event.type
        self._content = None
        if isinstance(event, MessageEvent):
            self._content = MessageChain(event.message.message)
        self.reply: Reply | None = None

    async def prompt(
        self,
        message: str | Iterable[str | Element],
        timeout: float = 120,
        timeout_message: str | Iterable[str | Element] = "等待超时",
        keep_sender: bool = True,
    ) -> MessageChain:
        """发送提示消息, 并等待回复

        Args:
            message: 要发送的消息
            timeout: 等待超时时间
            timeout_message: 超时后发送的消息
            keep_sender: 是否只允许原发送者回复
        """
        await self.send(message)

        async def waiter(content: MessageChain, session: Session[MessageEvent]):
            if self.event.channel:
                if self.event.channel.id == session.event.channel.id and (
                    not keep_sender
                    or (self.event.user and session.event.user and self.event.user.id == session.event.user.id)
                ):
                    return content
            elif self.event.user:
                if self.event.user.id == session.event.user.id:
                    return content

        waiter.__annotations__ = {"content": MessageChain, "session": self.__class__}

        step = StepOut([MessageEvent], waiter)

        result = await step.wait(timeout=timeout)
        if not result:
            await self.send(timeout_message)
            raise HandlerStop()
        return result

    def stop(self) -> NoReturn:
        raise HandlerStop()

    @property
    def quote(self):
        if isinstance(self.event, MessageEvent):
            return self.event.quote
        return None

    @property
    def user(self) -> User:
        if not self.event.user:
            raise RuntimeError(f"Event {self.event.type!r} has no User")
        return self.event.user

    @property
    def guild(self) -> Guild:
        if not self.event.guild:
            raise RuntimeError(f"Event {self.event.type!r} has no Guild")
        return self.event.guild

    @property
    def channel(self) -> Channel:
        if not self.event.channel:
            raise RuntimeError(f"Event {self.event.type!r} has no Channel")
        return self.event.channel

    @property
    def member(self) -> Member:
        if not self.event.member:
            raise RuntimeError(f"Event {self.event.type!r} has no Member")
        return self.event.member

    @property
    def content(self) -> str:
        if self._content:
            return str(self._content)
        raise RuntimeError(f"Event {self.event.type!r} has no Content")

    @content.setter
    def content(self, value: str):
        self._content = MessageChain(value)

    @property
    def elements(self) -> MessageChain:
        if self._content:
            return self._content
        raise RuntimeError(f"Event {self.event.type!r} has no Content")

    @elements.setter
    def elements(self, value: MessageChain):
        self._content = value

    async def _send(self, channel_id: str, message: str | Iterable[str | Element]):
        return await self.account.protocol.send_message(channel_id, message, self.event)

    async def send(
        self,
        message: str | Iterable[str | Element],
    ) -> list[MessageReceipt]:
        if not self.event._origin.channel:
            raise RuntimeError("Event cannot be replied to!")
        return await self.account.protocol.send_message(self.event._origin.channel.id, message, self.event)

    async def send_message(
        self,
        message: str | Iterable[str | Element],
    ) -> list[MessageReceipt]:
        """发送消息

        Args:
            message: 要发送的消息
        """
        if not self.event.channel:
            raise RuntimeError("Event cannot be replied to!")
        return await self.account.protocol.send_message(self.event.channel.id, message, self.event)

    async def send_private_message(
        self,
        message: str | Iterable[str | Element],
    ) -> list[MessageReceipt]:
        """发送私聊消息

        Args:
            message: 要发送的消息
        """
        channel = await self.user_channel_create()
        return await self.account.protocol.send_message(channel.id, message, self.event)

    async def update_message(
        self,
        message: str | Iterable[str | Element],
    ):
        """更新消息

        Args:
            message: 要更新的消息
        """
        if not self.event.channel:
            raise RuntimeError("Event cannot be replied to!")
        if not self.event.message:
            raise RuntimeError("Event cannot update message")
        return await self.account.protocol.update_message(self.event.channel, self.event.message.id, message)

    async def message_create(
        self,
        content: str,
    ) -> list[MessageReceipt]:
        if not self.event.channel:
            raise RuntimeError("Event cannot be replied to!")
        return await self.account.protocol.send_message(self.event.channel.id, content, self.event)

    async def message_delete(self, message_id: str) -> None:
        if not self.event.channel:
            raise RuntimeError("Event cannot be replied to!")
        await self.account.protocol.message_delete(
            self.event.channel.id,
            message_id,
        )

    async def message_update(
        self,
        content: str,
    ) -> None:
        if not self.event.channel:
            raise RuntimeError("Event cannot be replied to!")
        if not self.event.message:
            raise RuntimeError("Event cannot update message")
        await self.account.protocol.message_update(
            self.event.channel.id,
            self.event.message.id,
            content,
        )

    async def channel_create(self, data: Channel) -> Channel:
        if not self.event.guild:
            raise RuntimeError("Event cannot use to create channel!")
        return await self.account.protocol.channel_create(self.event.guild.id, data)

    async def channel_list(self, next_token: str | None = None) -> PageResult[Channel]:
        if not self.event.guild:
            raise RuntimeError("Event cannot use to list channel!")
        return await self.account.protocol.channel_list(self.event.guild.id, next_token)

    async def channel_update(
        self,
        data: Channel,
    ) -> None:
        if not self.event.channel:
            raise RuntimeError("Event cannot use to update channel!")
        return await self.account.protocol.channel_update(self.event.channel.id, data)

    async def channel_delete(self) -> None:
        if not self.event.channel:
            raise RuntimeError("Event cannot use to delete channel!")
        return await self.account.protocol.channel_delete(
            self.event.channel.id,
        )

    async def user_channel_create(self) -> Channel:
        if not self.event.user:
            raise RuntimeError("Event cannot use to create user channel!")
        return await self.account.protocol.user_channel_create(
            self.event.user.id, self.event.guild.id if self.event.guild else None
        )

    async def guild_member_list(self, next_token: str | None = None) -> PageResult[Member]:
        if not self.event.guild:
            raise RuntimeError("Event cannot use to list member!")
        return await self.account.protocol.guild_member_list(self.event.guild.id, next_token)

    async def guild_member_get(self, user_id: str | None = None) -> Member:
        if not self.event.guild:
            raise RuntimeError("Event cannot use to get member!")
        if user_id:
            return await self.account.protocol.guild_member_get(self.event.guild.id, user_id)
        if not self.event.user:
            raise RuntimeError("Event cannot use to get member!")
        return await self.account.protocol.guild_member_get(self.event.guild.id, self.event.user.id)

    async def guild_member_kick(self, user_id: str | None = None, permanent: bool = False) -> None:
        if not self.event.guild:
            raise RuntimeError("Event cannot use to kick member!")
        if user_id:
            return await self.account.protocol.guild_member_kick(self.event.guild.id, user_id, permanent)
        if not self.event.user:
            raise RuntimeError("Event cannot use to kick member!")
        return await self.account.protocol.guild_member_kick(self.event.guild.id, self.event.user.id, permanent)

    async def guild_member_role_set(self, role_id: str, user_id: str | None = None) -> None:
        if not self.event.guild:
            raise RuntimeError("Event cannot use to guild member role set!")
        if user_id:
            return await self.account.protocol.guild_member_role_set(self.event.guild.id, user_id, role_id)
        if not self.event.user:
            raise RuntimeError("Event cannot use to guild member role set!")
        return await self.account.protocol.guild_member_role_set(self.event.guild.id, self.event.user.id, role_id)

    async def guild_member_role_unset(self, role_id: str, user_id: str | None = None) -> None:
        if not self.event.guild:
            raise RuntimeError("Event cannot use to guild member role unset!")
        if user_id:
            return await self.account.protocol.guild_member_role_unset(self.event.guild.id, user_id, role_id)
        if not self.event.user:
            raise RuntimeError("Event cannot use to guild member role unset!")
        return await self.account.protocol.guild_member_role_unset(self.event.guild.id, self.event.user.id, role_id)

    async def guild_role_list(self, next_token: str | None = None) -> PageResult[Role]:
        if not self.event.guild:
            raise RuntimeError("Event cannot use to list role!")
        return await self.account.protocol.guild_role_list(self.event.guild.id, next_token)

    async def guild_role_create(
        self,
        role: Role,
    ) -> Role:
        if not self.event.guild:
            raise RuntimeError("Event cannot use to create role!")
        return await self.account.protocol.guild_role_create(self.event.guild.id, role)

    async def guild_role_update(
        self,
        role_id: str,
        role: Role,
    ) -> None:
        if not self.event.guild:
            raise RuntimeError("Event cannot use to update role!")
        return await self.account.protocol.guild_role_update(self.event.guild.id, role_id, role)

    async def guild_role_delete(self, role_id: str) -> None:
        if not self.event.guild:
            raise RuntimeError("Event cannot use to delete role!")
        return await self.account.protocol.guild_role_delete(
            self.event.guild.id,
            role_id,
        )

    async def reaction_create(
        self,
        emoji: str,
    ) -> None:
        if not self.event.channel:
            raise RuntimeError("Event cannot be replied to!")
        if not self.event.message:
            raise RuntimeError("Event cannot create reaction")
        return await self.account.protocol.reaction_create(self.event.channel.id, self.event.message.id, emoji)

    async def reaction_delete(
        self,
        emoji: str,
        user_id: str | None = None,
    ) -> None:
        if not self.event.channel:
            raise RuntimeError("Event cannot be replied to!")
        if not self.event.message:
            raise RuntimeError("Event cannot delete reaction")
        return await self.account.protocol.reaction_delete(self.event.channel.id, self.event.message.id, emoji, user_id)

    async def reaction_clear(
        self,
        emoji: str | None = None,
    ) -> None:
        if not self.event.channel:
            raise RuntimeError("Event cannot be replied to!")
        if not self.event.message:
            raise RuntimeError("Event cannot clear reaction")
        return await self.account.protocol.reaction_clear(
            self.event.channel.id,
            self.event.message.id,
            emoji,
        )

    async def reaction_list(
        self,
        emoji: str,
        next_token: str | None = None,
    ) -> PageResult[User]:
        if not self.event.channel:
            raise RuntimeError("Event cannot be replied to!")
        if not self.event.message:
            raise RuntimeError("Event cannot list reaction")
        return await self.account.protocol.reaction_list(
            self.event.channel.id, self.event.message.id, emoji, next_token
        )

    async def request_approve(
        self,
        approve: bool,
        comment: str,
    ):
        if isinstance(self.event, FriendRequestEvent):
            return await self.friend_approve(approve, comment)
        if isinstance(self.event, GuildRequestEvent):
            return await self.guild_approve(approve, comment)
        if isinstance(self.event, GuildMemberRequestEvent):
            return await self.guild_member_approve(approve, comment)
        raise RuntimeError("Event cannot approve request")

    async def friend_approve(
        self,
        approve: bool,
        comment: str,
    ):
        if not isinstance(self.event, FriendRequestEvent):
            raise RuntimeError("Event cannot approve friend request")
        return await self.account.protocol.friend_approve(self.event.message.id, approve, comment)

    async def guild_approve(
        self,
        approve: bool,
        comment: str,
    ):
        if not isinstance(self.event, GuildRequestEvent):
            raise RuntimeError("Event cannot approve guild request")
        return await self.account.protocol.guild_approve(self.event.message.id, approve, comment)

    async def guild_member_approve(
        self,
        approve: bool,
        comment: str,
    ):
        if not isinstance(self.event, GuildMemberRequestEvent):
            raise RuntimeError("Event cannot approve guild member request")
        return await self.account.protocol.guild_member_approve(self.event.message.id, approve, comment)



import asyncio
from typing import Callable, Optional, TypeVar, Union, cast, overload

from arclet.alconna import Alconna, Arg, Args, CommandMeta, Namespace, command_manager, config
from arclet.alconna.tools.construct import AlconnaString, alconna_from_format
from arclet.alconna.typing import TAValue
from arclet.letoderea import Provider, Scope, Subscriber, es
from arclet.letoderea.handler import generate_contexts
from arclet.letoderea.provider import ProviderFactory, get_providers
from arclet.letoderea.typing import Contexts, TTarget
from nepattern import DirectPattern
from satori.element import Text
from tarina.string import split
from tarina.trie import CharTrie

from ..event.base import MessageCreatedEvent
from ..event.command import CommandExecute
from ..event.config import ConfigReload
from ..message import MessageChain
from ..plugin import RootlessPlugin, _current_plugin
from ..session import Session
from .argv import MessageArgv  # noqa: F401
from .model import CommandResult, Match, Query
from .plugin import mount
from .provider import AlconnaProviderFactory, AlconnaSuppiler, MessageJudges

TM = TypeVar("TM", bound=Union[str, MessageChain])


def get_cmd(target: Subscriber):
    if sub := target.get_propagator(AlconnaSuppiler.supply):
        return sub.callable_target.__self__.cmd  # type: ignore
    raise ValueError("Subscriber has no command.")


class EntariCommands:
    __namespace__ = "Entari"

    def __init__(self, need_notice_me: bool = False, need_reply_me: bool = False, use_config_prefix: bool = True):
        self.trie: CharTrie[str] = CharTrie()
        self.scope = Scope("entari.command")
        self.scope.bind(*get_providers(MessageCreatedEvent), AlconnaProviderFactory())
        self.judge = MessageJudges(need_notice_me, need_reply_me, use_config_prefix)
        config.namespaces["Entari"] = Namespace(
            self.__namespace__,
            to_text=lambda x: x.text if x.__class__ is Text else None,
            converter=lambda x: MessageChain(x),
        )
        es.on(CommandExecute, self.execute)

    @property
    def all_helps(self) -> str:
        return command_manager.all_command_help(namespace=self.__namespace__)

    def get_help(self, command: str) -> str:
        return command_manager.get_command(f"{self.__namespace__}::{command}").get_help()

    async def handle(self, session: Session, message: MessageChain, ctx: Contexts):
        msg = str(message).lstrip()
        if not msg:
            return
        if matches := list(self.trie.prefixes(msg)):
            subs = [self.scope.subscribers[res.value][0] for res in matches if res.value in self.scope.subscribers]
            results = await asyncio.gather(*(sub.handle(ctx.copy()) for sub in subs))
            for result in results:
                if result is not None:
                    await session.send(result)
            return
        # shortcut
        data = split(msg, " ")
        for value in self.trie.values():
            if value not in self.scope.subscribers:
                continue
            sub = self.scope.subscribers[value][0]
            try:
                command_manager.find_shortcut(get_cmd(sub), data)
            except ValueError:
                continue
            result = await sub.handle(ctx.copy())
            if result is not None:
                await session.send(result)

    async def execute(self, event: CommandExecute):
        ctx = await generate_contexts(event)
        msg = str(event.command)
        if matches := list(self.trie.prefixes(msg)):
            subs = [self.scope.subscribers[res.value][0] for res in matches if res.value in self.scope.subscribers]
            results = await asyncio.gather(*(sub.handle(ctx.copy(), inner=True) for sub in subs))
            for result in results:
                if result is not None:
                    return result
        data = split(msg, " ")
        for value in self.trie.values():
            if value not in self.scope.subscribers:
                continue
            sub = self.scope.subscribers[value][0]
            try:
                command_manager.find_shortcut(get_cmd(sub), data)
            except ValueError:
                continue
            result = await sub.handle(ctx.copy(), inner=True)
            if result is not None:
                return result

    def command(
        self,
        command: str,
        help_text: Optional[str] = None,
        providers: Optional[list[Union[Provider, type[Provider], ProviderFactory, type[ProviderFactory]]]] = None,
    ):
        class Command(AlconnaString):
            def __call__(_cmd_self, func: TTarget[Optional[TM]]) -> Subscriber[Optional[TM]]:
                return self.on(_cmd_self.build(), providers)(func)

        return Command(command, help_text)

    @overload
    def on(
        self,
        command: Alconna,
        providers: Optional[list[Union[Provider, type[Provider], ProviderFactory, type[ProviderFactory]]]] = None,
    ) -> Callable[[TTarget[Optional[TM]]], Subscriber[Optional[TM]]]: ...

    @overload
    def on(
        self,
        command: str,
        providers: Optional[list[Union[Provider, type[Provider], ProviderFactory, type[ProviderFactory]]]] = None,
        *,
        args: Optional[dict[str, Union[TAValue, Args, Arg]]] = None,
        meta: Optional[CommandMeta] = None,
    ) -> Callable[[TTarget[Optional[TM]]], Subscriber[Optional[TM]]]: ...

    def on(
        self,
        command: Union[Alconna, str],
        providers: Optional[list[Union[Provider, type[Provider], ProviderFactory, type[ProviderFactory]]]] = None,
        *,
        args: Optional[dict[str, Union[TAValue, Args, Arg]]] = None,
        meta: Optional[CommandMeta] = None,
    ) -> Callable[[TTarget[Optional[TM]]], Subscriber[Optional[TM]]]:
        plg = _current_plugin.get()
        providers = providers or []

        def wrapper(func: TTarget[Optional[TM]]) -> Subscriber[Optional[TM]]:
            if isinstance(command, str):
                mapping = {arg.name: arg.value for arg in Args.from_callable(func)[0]}
                mapping.update(args or {})  # type: ignore
                _command = alconna_from_format(command, mapping, meta, union=False)
                _command.reset_namespace(self.__namespace__)
                key = _command.name + "".join(
                    f" {arg.value.target}" for arg in _command.args if isinstance(arg.value, DirectPattern)
                )
                target = self.scope.register(func, providers=providers)
                target.propagate(AlconnaSuppiler(_command))
                if plg:
                    target.propagates(*plg._scope.propagators)
                self.trie[key] = target.id

                def _remove(_):
                    self.scope.remove_subscriber(_)
                    command_manager.delete(get_cmd(_))
                    self.trie.pop(key, None)  # type: ignore

                if plg:
                    plg.collect(lambda: _remove(target))
                    plg._extra.setdefault("commands", []).append(([], _command.command))
                else:
                    target._dispose = _remove
                return target

            _command = cast(Alconna, command)
            if not isinstance(command.command, str):
                raise TypeError("Command name must be a string.")
            _command.reset_namespace(self.__namespace__)
            keys = []
            if not _command.prefixes:
                keys.append(_command.command)
            elif not all(isinstance(i, str) for i in _command.prefixes):
                raise TypeError("Command prefixes must be a list of string.")
            else:
                for prefix in cast(list[str], _command.prefixes):
                    keys.append(prefix + _command.command)

            target = self.scope.register(func, providers=providers)
            target.propagate(AlconnaSuppiler(_command))
            if plg:
                target.propagates(*plg._scope.propagators)

            for _key in keys:
                self.trie[_key] = target.id

            def _remove(_):
                self.scope.remove_subscriber(_)
                command_manager.delete(get_cmd(_))
                for _key in keys:
                    self.trie.pop(_key, None)  # type: ignore

            if plg:
                plg._extra.setdefault("commands", []).append((_command.prefixes, _command.command))
                plg.collect(lambda: _remove(target))
            else:
                target._dispose = _remove
            return target

        return wrapper


_commands = EntariCommands()


def config_commands(need_notice_me: bool = False, need_reply_me: bool = False, use_config_prefix: bool = True):
    _commands.judge.need_notice_me = need_notice_me
    _commands.judge.need_reply_me = need_reply_me
    _commands.judge.use_config_prefix = use_config_prefix


command = _commands.command
on = _commands.on


async def execute(message: Union[str, MessageChain]):
    res = await es.post(CommandExecute(message))
    if res:
        return res.value


@RootlessPlugin.apply("commands")
def _(plg: RootlessPlugin):
    if "need_notice_me" in plg.config:
        _commands.judge.need_notice_me = plg.config["need_notice_me"]
    if "need_reply_me" in plg.config:
        _commands.judge.need_reply_me = plg.config["need_reply_me"]
    if "use_config_prefix" in plg.config:
        _commands.judge.use_config_prefix = plg.config["use_config_prefix"]

    plg.dispatch(MessageCreatedEvent).handle(_commands.handle).propagate(_commands.judge)

    @plg.use(ConfigReload)
    def update(event: ConfigReload):
        if event.scope != "plugin":
            return
        if event.key != ".commands":
            return
        if "need_notice_me" in event.value:
            _commands.judge.need_notice_me = event.value["need_notice_me"]
        if "need_reply_me" in event.value:
            _commands.judge.need_reply_me = event.value["need_reply_me"]
        if "use_config_prefix" in event.value:
            _commands.judge.use_config_prefix = event.value["use_config_prefix"]
        return True


__all__ = ["_commands", "config_commands", "Match", "Query", "execute", "CommandResult", "mount", "command", "on"]



import asyncio
from collections.abc import Awaitable
from datetime import datetime
from typing import Callable, Final, Optional, Union
from typing_extensions import TypeAlias

from arclet.letoderea import STOP, Propagator
from arclet.letoderea.typing import run_sync
from tarina import is_coroutinefunction

from . import common, message
from ..message import MessageChain
from ..session import Session
from .common import parse as parse

_SessionFilter: TypeAlias = Union[Callable[[Session], bool], Callable[[Session], Awaitable[bool]]]


class _Filter:
    user = staticmethod(common.user)
    guild = staticmethod(common.guild)
    channel = staticmethod(common.channel)
    self_ = staticmethod(common.account)
    platform = staticmethod(common.platform)
    direct = staticmethod(message.direct_message)
    private = staticmethod(message.direct_message)
    direct_message = staticmethod(message.direct_message)
    public = staticmethod(message.public_message)
    public_message = staticmethod(message.public_message)
    notice_me = staticmethod(message.notice_me)
    reply_me = staticmethod(message.reply_me)
    to_me = staticmethod(message.to_me)

    def __call__(self, func: _SessionFilter):
        _func = func if is_coroutinefunction(func) else run_sync(func)

        async def _(session: Session):
            if not await _func(session):  # type: ignore
                return STOP

        return _


filter_: Final[_Filter] = _Filter()
F = filter_


class Interval(Propagator):
    def __init__(self, interval: float, limit_prompt: Optional[Union[str, MessageChain]] = None):
        self.success = True
        self.last_time = None
        self.interval = interval
        self.limit_prompt = limit_prompt

    async def before(self, session: Optional[Session] = None):
        if not self.last_time:
            return
        self.success = (datetime.now() - self.last_time).total_seconds() > self.interval
        if not self.success:
            if session and self.limit_prompt:
                await session.send(self.limit_prompt)
            return STOP

    async def after(self):
        if self.success:
            self.last_time = datetime.now()

    def compose(self):
        yield self.before, True, 15
        yield self.after, False, 60


class Semaphore(Propagator):
    def __init__(self, count: int, limit_prompt: Optional[Union[str, MessageChain]] = None):
        self.count = count
        self.limit_prompt = limit_prompt
        self.semaphore = asyncio.Semaphore(count)

    async def before(self, session: Optional[Session] = None):
        if not await self.semaphore.acquire():
            if session and self.limit_prompt:
                await session.send(self.limit_prompt)
            return STOP

    async def after(self):
        self.semaphore.release()

    def compose(self):
        yield self.before, True, 15
        yield self.after, False, 60



from __future__ import annotations

from dataclasses import dataclass
from dataclasses import field as field
from dataclasses import fields, is_dataclass
from inspect import Signature
import json
import os
from pathlib import Path
import re
from typing import Any, Callable, ClassVar, TypedDict, TypeVar, get_args, get_origin
from typing_extensions import dataclass_transform
import warnings

_available_dc_attrs = set(Signature.from_callable(dataclass).parameters.keys())

ENV_CONTEXT_PAT = re.compile(r"\$\{\{\s?env\.(?P<name>[^}]+)\s?\}\}")


class BasicConfig(TypedDict, total=False):
    network: list[dict[str, Any]]
    ignore_self_message: bool
    skip_req_missing: bool
    log_level: int | str
    prefix: list[str]
    cmd_count: int


@dataclass
class EntariConfig:
    path: Path
    basic: BasicConfig = field(default_factory=dict, init=False)  # type: ignore
    plugin: dict[str, dict] = field(default_factory=dict, init=False)
    prelude_plugin: list[str] = field(default_factory=list, init=False)
    plugin_extra_files: list[str] = field(default_factory=list, init=False)
    updater: Callable[[EntariConfig], None]

    instance: ClassVar[EntariConfig]

    def __post_init__(self):
        self.__class__.instance = self
        self.reload()

    @staticmethod
    def _load_plugin(path: Path):
        if path.suffix.startswith(".json"):
            with path.open("r", encoding="utf-8") as f:
                return json.load(f)
        if path.suffix in (".yaml", ".yml"):
            try:
                import yaml
            except ImportError:
                raise RuntimeError("yaml is not installed")

            with path.open("r", encoding="utf-8") as f:
                return yaml.safe_load(f)
        raise NotImplementedError(f"unsupported plugin config file format: {path!s}")

    def reload(self):
        self.updater(self)
        self.plugin_extra_files: list[str] = self.plugin.pop("$files", [])  # type: ignore
        for file in self.plugin_extra_files:
            path = Path(file)
            if not path.exists():
                raise FileNotFoundError(file)
            if path.is_dir():
                for _path in path.iterdir():
                    if not _path.is_file():
                        continue
                    self.plugin[_path.stem] = self._load_plugin(_path)
            else:
                self.plugin[path.stem] = self._load_plugin(path)

        self.plugin.setdefault(".commands", {})
        self.prelude_plugin = self.plugin.pop("$prelude", [])  # type: ignore
        disabled = []
        for k, v in self.plugin.items():
            if v is True:
                self.plugin[k] = {}
                warnings.warn(
                    f"`True` usage in plugin '{k}' config is deprecated, use empty dict instead", DeprecationWarning
                )
            elif v is False:
                disabled.append(k)
        for k in disabled:
            self.plugin[f"~{k}"] = self.plugin.pop(k)
            warnings.warn(
                f"`False` usage in plugin '{k}' config is deprecated, use `~` prefix instead", DeprecationWarning
            )

    @classmethod
    def load(cls, path: str | os.PathLike[str] | None = None) -> EntariConfig:
        try:
            import dotenv

            dotenv.load_dotenv()
        except ImportError:
            dotenv = None  # noqa
            pass
        if path is None:
            if "ENTARI_CONFIG_FILE" in os.environ:
                _path = Path(os.environ["ENTARI_CONFIG_FILE"])
            elif (Path.cwd() / ".entari.json").exists():
                _path = Path.cwd() / ".entari.json"
            else:
                _path = Path.cwd() / "entari.yml"
        else:
            _path = Path(path)
        if not _path.exists():
            return cls(_path, lambda _: None)
        if not _path.is_file():
            raise ValueError(f"{_path} is not a file")

        if _path.suffix.startswith(".json"):

            def _updater(self: EntariConfig):
                with self.path.open("r", encoding="utf-8") as f:
                    data = json.load(f)
                    if "entari" in data:
                        data = data["entari"]
                    self.basic = data.get("basic", {})
                    self.plugin = data.get("plugins", {})

            obj = cls(_path, _updater)
            cls.instance = obj
            return obj
        if _path.suffix in (".yaml", ".yml"):
            try:
                import yaml
            except ImportError:
                raise RuntimeError("yaml is not installed. Please install with `arclet-entari[yaml]`")

            def _updater(self: EntariConfig):
                with self.path.open("r", encoding="utf-8") as f:
                    text = f.read()
                    text = ENV_CONTEXT_PAT.sub(lambda m: os.environ.get(m["name"], ""), text)
                    data = yaml.safe_load(text)
                    if "entari" in data:
                        data = data["entari"]
                    self.basic = data.get("basic", {})
                    self.plugin = data.get("plugins", {})

            return cls(_path, _updater)
        raise NotImplementedError(f"unsupported config file format: {_path!s}")


load_config = EntariConfig.load


_config_model_validators = {}

C = TypeVar("C")


def config_validator_register(base: type):
    def wrapper(func: Callable[[dict[str, Any], type[C]], C]):
        _config_model_validators[base] = func
        return func

    return wrapper


def config_model_validate(base: type[C], data: dict[str, Any]) -> C:
    for b in base.__mro__[-2::-1]:
        if b in _config_model_validators:
            return _config_model_validators[b](data, base)
    return base(**data)


@dataclass_transform(kw_only_default=True)
class BasicConfModel:
    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        dataclass(**{k: v for k, v in kwargs.items() if k in _available_dc_attrs})(cls)


@config_validator_register(BasicConfModel)
def _basic_config_validate(data: dict[str, Any], base: type[C]) -> C:
    def _nested_validate(namespace: dict[str, Any], cls):
        result = {}
        for field_ in fields(cls):
            if field_.name not in namespace:
                continue
            if is_dataclass(field_.type):
                result[field_.name] = _nested_validate(namespace[field_.name], field_.type)
            elif get_origin(field_.type) is list and is_dataclass(get_args(field_.type)[0]):
                result[field_.name] = [_nested_validate(d, get_args(field_.type)[0]) for d in namespace[field_.name]]
            elif get_origin(field_.type) is set and is_dataclass(get_args(field_.type)[0]):
                result[field_.name] = {_nested_validate(d, get_args(field_.type)[0]) for d in namespace[field_.name]}
            elif get_origin(field_.type) is dict and is_dataclass(get_args(field_.type)[1]):
                result[field_.name] = {
                    k: _nested_validate(v, get_args(field_.type)[1]) for k, v in namespace[field_.name].items()
                }
            elif get_origin(field_.type) is tuple:
                args = get_args(field_.type)
                result[field_.name] = tuple(
                    _nested_validate(d, args[i]) if is_dataclass(args[i]) else d
                    for i, d in enumerate(namespace[field_.name])
                )
            else:
                result[field_.name] = namespace[field_.name]
        return cls(**result)

    return _nested_validate(data, base)



from importlib.util import find_spec
from pathlib import Path

from arclet.alconna import Alconna, Args, CommandMeta, MultiVar, Option, Subcommand, command_manager
from arclet.alconna.tools.formatter import RichConsoleFormatter

from arclet.entari import Entari

alc = Alconna(
    "entari",
    Subcommand(
        "new",
        Option("-d|--dev", help_text="是否生成开发用配置文件"),
        Option("-P|--plugins", Args["names/", MultiVar(str)], help_text="指定增加哪些插件"),
        help_text="新建一个 Entari 配置文件",
    ),
    Subcommand("run", help_text="运行 Entari"),
    Option("-c|--config", Args["path/", str], help_text="指定配置文件路径"),
    meta=CommandMeta(
        "Entari App Launcher",
    ),
    formatter_type=RichConsoleFormatter,
)


JSON_BASIC_TEMPLATE = """\
{
  "basic": {
    "network": [
      {
        "type": "websocket",
        "host": "127.0.0.1",
        "port": 5140,
        "path": "satori"
      }
    ],
    "ignore_self_message": true,
    "log_level": "info",
    "prefix": ["/"]
  },
"""

JSON_PLUGIN_BLANK_TEMPLATE = """\
  "plugins": {{
{plugins}
  }}
}}
"""

JSON_PLUGIN_COMMON_TEMPLATE = """\
  "plugins": {
    ".record_message": {},
    "::echo": {},
    "::inspect": {}
  }
}
"""

JSON_PLUGIN_DEV_TEMPLATE = """\
  "plugins": {
    "$prelude": [
      "::auto_reload"
    ],
    ".record_message": {
      "record_send": true,
    },
    "::help": {},
    "::echo": {},
    "::inspect": {},
    "::auto_reload": {
      "watch_config": true
    }
  }
}
"""


YAML_BASIC_TEMPLATE = """\
basic:
  network:
    - type: websocket
      host: "127.0.0.1"
      port: 5140
      path: "satori"
  ignore_self_message: true
  log_level: "info"
  prefix: ["/"]
"""

YAML_PLUGIN_BLANK_TEMPLATE = """\
plugins:
{plugins}
"""


YAML_PLUGIN_COMMON_TEMPLATE = """\
plugins:
  .record_message: {}
  ::echo: {}
  ::inspect: {}
"""

YAML_PLUGIN_DEV_TEMPLATE = """\
plugins:
  $prelude:
    - ::auto_reload
  .record_message:
    record_send: true
  ::echo: {}
  ::help: {}
  ::inspect: {}
  ::auto_reload:
    watch_config: true
"""


def main():
    res = alc()
    if (not res.matched or res.non_component) and not res.error_info:
        print(alc.get_help())
        return
    if res.find("new"):
        is_dev = res.find("new.dev")
        names = res.query[tuple[str, ...]]("new.plugins.names", ())
        if (path := res.query[str]("config.path", None)) is None:
            if find_spec("yaml"):
                _path = Path.cwd() / "entari.yml"
            else:
                _path = Path.cwd() / ".entari.json"
        else:
            _path = Path(path)
        if _path.exists():
            print(f"{_path} already exists")
            return
        if _path.suffix.startswith(".json"):
            if names:
                PT = JSON_PLUGIN_BLANK_TEMPLATE.format(plugins=",\n".join(f'    "{name}": {{}}' for name in names))
            elif is_dev:
                PT = JSON_PLUGIN_DEV_TEMPLATE
            else:
                PT = JSON_PLUGIN_COMMON_TEMPLATE

            with _path.open("w", encoding="utf-8") as f:
                f.write(JSON_BASIC_TEMPLATE + PT)
            print(f"Config file created at {_path}")
            return
        if _path.suffix in (".yaml", ".yml"):
            if names:
                PT = YAML_PLUGIN_BLANK_TEMPLATE.format(plugins="\n".join(f"  {name}: {{}}" for name in names))
            elif is_dev:
                PT = YAML_PLUGIN_DEV_TEMPLATE
            else:
                PT = YAML_PLUGIN_COMMON_TEMPLATE

            with _path.open("w", encoding="utf-8") as f:
                f.write(YAML_BASIC_TEMPLATE + PT)
            print(f"Config file created at {_path}")
            return
        print(f"Unsupported file extension: {_path.suffix}")
        return
    if res.find("run"):
        command_manager.delete(alc)
        entari = Entari.load(res.query[str]("config.path", None))
        entari.run()



from dataclasses import dataclass
from typing import Optional

from arclet.letoderea import make_event


@dataclass
@make_event(name="entari.event/plugin/loaded_success")
class PluginLoadedSuccess:
    name: str


@dataclass
@make_event(name="entari.event/plugin/loaded_failed")
class PluginLoadedFailed:
    name: str
    error: Optional[Exception] = None
    """若没有异常信息，说明该插件加载失败的原因是插件不存在。"""


@dataclass
@make_event(name="entari.event/plugin/unloaded")
class PluginUnloaded:
    name: str



from __future__ import annotations

import asyncio
from contextlib import suppress
import os

from arclet.alconna import config as alconna_config
from arclet.letoderea import Contexts, Param, Provider, ProviderFactory, es, global_providers
from creart import it
from launart import Launart, Service
from satori import LoginStatus
from satori.client import App
from satori.client.account import Account
from satori.client.config import Config, WebhookInfo, WebsocketsInfo
from satori.client.protocol import ApiProtocol
from satori.model import Event
from tarina.generic import get_origin

from .config import EntariConfig
from .event.base import MessageCreatedEvent, event_parse
from .event.config import ConfigReload
from .event.lifespan import AccountUpdate
from .event.send import SendResponse
from .logger import log
from .plugin import load_plugin, plugin_config, requires
from .plugin.model import RootlessPlugin
from .plugin.service import plugin_service
from .session import EntariProtocol, Session


class ApiProtocolProvider(Provider[ApiProtocol]):
    async def __call__(self, context: Contexts):
        if account := context.get("account"):
            return account.protocol


class SessionProvider(Provider[Session]):
    def validate(self, param: Param):
        return (get_origin(param.annotation) == Session) or super().validate(param)

    async def __call__(self, context: Contexts):
        if "session" in context and isinstance(context["session"], Session):
            return context["session"]
        if "$origin_event" in context and "account" in context:
            session = Session(context["account"], context["$event"])
            if "$message_content" in context:
                session.elements = context["$message_content"]
            if "$message_reply" in context:
                session.reply = context["$message_reply"]
            return session


class AccountProvider(Provider[Account]):
    async def __call__(self, context: Contexts):
        if "account" in context:
            return context["account"]


#
# class PluginProvider(Provider[Plugin]):
#     async def __call__(self, context: Contexts):
#         subscriber: Subscriber = context["$subscriber"]
#         func = subscriber.callable_target
#         if hasattr(func, "__globals__") and "__plugin__" in func.__globals__:  # type: ignore
#             return func.__globals__["__plugin__"]
#         if hasattr(func, "__module__"):
#             return plugin_service.plugins.get(func.__module__)


global_providers.extend([ApiProtocolProvider(), SessionProvider(), AccountProvider()])


@RootlessPlugin.apply("record_message")
def record(plg: RootlessPlugin):
    cfg = plugin_config()

    @plg.dispatch(MessageCreatedEvent).on(priority=0)
    async def log_msg(event: MessageCreatedEvent):
        log.message.info(
            f"[{event.channel.name or event.channel.id}] "
            f"{event.member.nick if event.member else (event.user.name or event.user.id)}"
            f"({event.user.id}) -> {event.message.content!r}"
        )

    if cfg.get("record_send", False):

        @plg.use(SendResponse)
        async def log_send(event: SendResponse):
            if event.session:
                log.message.info(
                    f"[{event.session.channel.name or event.session.channel.id}] <- {event.message.display()!r}"
                )
            else:
                log.message.info(f"[{event.channel}] <- {event.message.display()!r}")


class Entari(App):
    id = "entari.service"

    @classmethod
    def load(cls, path: str | os.PathLike[str] | None = None):
        return cls.from_config(EntariConfig.load(path))

    @classmethod
    def from_config(cls, config: EntariConfig | None = None):
        if not config:
            config = EntariConfig.instance
        ignore_self_message = config.basic.get("ignore_self_message", True)
        log_level = config.basic.get("log_level", "INFO")
        skip_req_missing = config.basic.get("skip_req_missing", False)
        configs = []
        for conf in config.basic.get("network", []):
            if conf["type"] in ("websocket", "websockets", "ws"):
                configs.append(WebsocketsInfo(**{k: v for k, v in conf.items() if k != "type"}))
            elif conf["type"] in ("webhook", "wh", "http"):
                configs.append(WebhookInfo(**{k: v for k, v in conf.items() if k != "type"}))
        return cls(
            *configs,
            log_level=log_level,
            ignore_self_message=ignore_self_message,
            skip_req_missing=skip_req_missing,
        )

    def __init__(
        self,
        *configs: Config,
        log_level: str | int = "INFO",
        ignore_self_message: bool = True,
        skip_req_missing: bool = False,
    ):
        from . import __version__

        es.global_skip_req_missing = skip_req_missing
        log.core.info(f"Entari <b><c>version {__version__}</c></b>")
        super().__init__(*configs, default_api_cls=EntariProtocol)
        if not hasattr(EntariConfig, "instance"):
            EntariConfig.load()
        alconna_config.command_max_count = EntariConfig.instance.basic.get("cmd_count", 4096)
        log.set_level(log_level)
        log.core.debug(f"Log level set to <y><c>{log_level}</c></y>")
        requires(*EntariConfig.instance.prelude_plugin)
        for plug in EntariConfig.instance.prelude_plugin:
            load_plugin(plug, prelude=True)
        plugins = [
            plug for plug in EntariConfig.instance.plugin if not plug.startswith("~") and not plug.startswith("$")
        ]
        requires(*plugins)
        for plug in plugins:
            load_plugin(plug)
        self.ignore_self_message = ignore_self_message
        self.register(self.handle_event)
        self.lifecycle(self.account_hook)
        self._ref_tasks = set()

        es.on(ConfigReload, self.reset_self)

    def reset_self(self, scope, key, value):
        if scope != "basic":
            return
        if key == "log_level":
            log.set_level(value)
            log.core.debug(f"Log level set to <y><c>{value}</c></y>")
        elif key == "ignore_self_message":
            self.ignore_self_message = value
        elif key == "network":
            for conn in self.connections:
                it(Launart).remove_component(conn)
            self.connections.clear()
            for conf in value:
                if conf["type"] in ("websocket", "websockets", "ws"):
                    self.apply(WebsocketsInfo(**{k: v for k, v in conf.items() if k != "type"}))
                elif conf["type"] in ("webhook", "wh", "http"):
                    self.apply(WebhookInfo(**{k: v for k, v in conf.items() if k != "type"}))
            for conn in self.connections:
                it(Launart).add_component(conn)

    def on_message(self, priority: int = 16):
        return es.on(MessageCreatedEvent, priority=priority)

    def ensure_manager(self, manager: Launart):
        self.manager = manager
        manager.add_component(plugin_service)

    async def handle_event(self, account: Account, event: Event):
        with suppress(NotImplementedError):
            ev = event_parse(account, event)
            if self.ignore_self_message and isinstance(ev, MessageCreatedEvent) and ev.user.id == account.self_id:
                return
            es.publish(ev)
            return

        log.core.warning(f"received unsupported event {event.type}: {event}")

    async def account_hook(self, account: Account, state: LoginStatus):
        es.publish(AccountUpdate(account, state))

    @classmethod
    def current(cls):
        return it(Launart).get_component(cls)


class EntariProvider(Provider[Entari]):
    priority = 1

    async def __call__(self, context: Contexts):
        return Entari.current()


class LaunartProvider(Provider[Launart]):
    priority = 10

    async def __call__(self, context: Contexts):
        return it(Launart)


class ServiceProviderFactory(ProviderFactory):
    priority = 10

    class _Provider(Provider[Service]):
        def __init__(self, origin: type[Service]):
            super().__init__()
            self.origin = origin

        def validate(self, param: Param):
            anno = get_origin(param.annotation)
            return isinstance(anno, type) and issubclass(anno, self.origin)

        async def __call__(self, context: Contexts):
            return it(Launart).get_component(self.origin)

    def validate(self, param: Param):
        anno = get_origin(param.annotation)
        if isinstance(anno, type) and issubclass(anno, Service):
            return self._Provider(anno)


global_providers.extend([EntariProvider(), LaunartProvider(), ServiceProviderFactory()])  # type: ignore

es.loop = it(asyncio.AbstractEventLoop)



from arclet.letoderea import STOP
from satori import ChannelType

from ..session import Session


async def direct_message(sess: Session):
    if sess.channel.type != ChannelType.DIRECT:
        return STOP


async def public_message(sess: Session):
    if sess.channel.type == ChannelType.DIRECT:
        return STOP


def reply_me(is_reply_me: bool = False):
    if not is_reply_me:
        return STOP


def notice_me(is_notice_me: bool = False):
    if not is_notice_me:
        return STOP


def to_me(is_reply_me: bool = False, is_notice_me: bool = False):
    if not is_reply_me and not is_notice_me:
        return STOP



from __future__ import annotations

from typing import Any

from arclet.alconna import Alconna, command_manager
from arclet.letoderea import Provider, ProviderFactory, es

from ..event import MessageCreatedEvent
from ..event.command import CommandExecute
from ..plugin.model import Plugin, PluginDispatcher
from .model import Match, Query
from .provider import AlconnaProviderFactory, AlconnaSuppiler, Assign, MessageJudges, _seminal

exec_pub = es.define(CommandExecute)
exec_pub.bind(AlconnaProviderFactory())


class AlconnaPluginDispatcher(PluginDispatcher):

    def __init__(
        self,
        plugin: Plugin,
        command: Alconna,
        need_reply_me: bool = False,
        need_notice_me: bool = False,
        use_config_prefix: bool = True,
    ):
        plugin._extra.setdefault("commands", []).append((command.prefixes, command.command))
        self.supplier = AlconnaSuppiler(command)
        super().__init__(plugin, MessageCreatedEvent, command.path)
        self.propagators.append(
            MessageJudges(need_reply_me, need_notice_me, use_config_prefix),
        )
        self.propagators.append(self.supplier)
        self.providers.append(AlconnaProviderFactory())

        @plugin.collect
        def dispose():
            command_manager.delete(self.supplier.cmd)
            del self.supplier.cmd
            del self.supplier

    def assign(
        self,
        path: str,
        value: Any = _seminal,
        or_not: bool = False,
        priority: int = 16,
        providers: list[Provider | type[Provider] | ProviderFactory | type[ProviderFactory]] | None = None,
    ):
        assign = Assign(path, value, or_not)
        wrapper = self.register(priority=priority, providers=providers)

        def decorator(func):
            sub = wrapper(func)
            sub.propagate(assign)
            return sub

        return decorator

    def on_execute(
        self,
        priority: int = 16,
        providers: list[Provider | type[Provider] | ProviderFactory | type[ProviderFactory]] | None = None,
    ):
        wrapper = self.plugin._scope.register(priority=priority, providers=providers, publisher=exec_pub)

        def decorator(func):
            sub = wrapper(func)
            sub.propagate(self.supplier)
            return sub

        return decorator

    Match = Match
    Query = Query


def mount(
    cmd: Alconna,
    need_reply_me: bool = False,
    need_notice_me: bool = False,
    use_config_prefix: bool = True,
) -> AlconnaPluginDispatcher:
    if not (plugin := Plugin.current()):
        raise LookupError("no plugin context found")
    return AlconnaPluginDispatcher(plugin, cmd, need_reply_me, need_notice_me, use_config_prefix)



from dataclasses import dataclass
from typing import Any, Optional, overload

from arclet.letoderea import make_event

from ..config import C, config_model_validate


@dataclass
@make_event(name="entari.event/config/reload")
class ConfigReload:
    scope: str
    key: str
    value: Any
    old: Optional[Any] = None

    __result_type__: type[bool] = bool

    @overload
    def plugin_config(self) -> dict[str, Any]: ...

    @overload
    def plugin_config(self, model_type: type[C]) -> C: ...

    def plugin_config(self, model_type: Optional[type[C]] = None):
        if self.scope != "plugin":
            raise ValueError("not a plugin config")
        if model_type:
            return config_model_validate(model_type, self.value)
        return self.value



from dataclasses import field
import random
from typing import Optional

from arclet.alconna import (
    Alconna,
    Args,
    Arparma,
    CommandMeta,
    Field,
    Option,
    Subcommand,
    SubcommandResult,
    command_manager,
    namespace,
    store_true,
)
from tarina import lang

from arclet.entari import BasicConfModel, Session, command, metadata, plugin_config


class Config(BasicConfModel):
    help_command: str = "help"
    help_alias: list[str] = field(default_factory=lambda: ["帮助", "命令帮助"])
    help_all_alias: list[str] = field(default_factory=lambda: ["所有帮助", "所有命令帮助"])
    page_size: Optional[int] = None


config = plugin_config(Config)


metadata(
    "help",
    ["RF-Tar-Railt <rf_tar_railt@qq.com>"],
    description="展示所有命令帮助",
    config=Config,
)


with namespace("builtin/help") as ns:
    ns.disable_builtin_options = {"shortcut"}

    help_cmd = Alconna(
        config.help_command,
        Args[
            "query#选择某条命令的id或者名称查看具体帮助;/?",
            str,
            Field(
                "-1",
                completion=lambda: f"试试 {random.randint(0, len(command_manager.get_commands()))}",
                unmatch_tips=lambda x: f"预期输入为某个命令的id或者名称，而不是 {x}\n例如：/帮助 0",
            ),
        ],
        Option(
            "--page",
            Args["index", int],
            help_text="查看指定页数的命令帮助",
        ),
        Subcommand(
            "--namespace",
            Args["target?;#指定的命名空间", str],
            Option("--list", help_text="列出所有命名空间", action=store_true, default=False),
            alias=["-N", "命名空间"],
            help_text="是否列出命令所属命名空间",
        ),
        Option("--hide", alias=["-H", "隐藏"], help_text="是否列出隐藏命令", action=store_true, default=False),
        meta=CommandMeta(
            description="显示所有命令帮助",
            usage="可以使用 --hide 参数来显示隐藏命令，使用 -P 参数来显示命令所属插件名称",
            example=f"${config.help_command} 1",
        ),
    )

    for alias in set(config.help_alias):
        help_cmd.shortcut(alias, {"prefix": True, "fuzzy": False})
    for alias in set(config.help_all_alias):
        help_cmd.shortcut(alias, {"args": ["--hide"], "prefix": True, "fuzzy": False})


def help_cmd_handle(arp: Arparma, interactive: bool = False):
    is_namespace = arp.query[SubcommandResult]("namespace")
    page = arp.query[int]("page.index", 1)
    target_namespace = is_namespace.args.get("target") if is_namespace else None
    cmds = [
        i
        for i in command_manager.get_commands(target_namespace or "")
        if not i.meta.hide or arp.query[bool]("hide.value", False)
    ]
    if is_namespace and is_namespace.options["list"].value and not target_namespace:
        namespaces = {i.namespace: 0 for i in cmds}
        return "\n".join(
            f" 【{str(index).rjust(len(str(len(namespaces))), '0')}】{n}" for index, n in enumerate(namespaces.keys())
        )

    help_names = set()
    for i in cmds:
        help_names.update(i.namespace_config.builtin_option_name["help"])

    footer = lang.require("manager", "help_footer").format(help="|".join(sorted(help_names, key=lambda x: len(x))))
    show_namespace = is_namespace and not is_namespace.options["list"].value and not target_namespace
    if (query := arp.all_matched_args["query"]) != "-1":
        if query.isdigit():
            index = int(query)
            if index < 0 or index >= len(cmds):
                return "查询失败！"
            slot = cmds[index]
        elif not (slot := next((i for i in cmds if query == i.command), None)):
            command_string = "\n".join(
                (
                    f"【{str(index).rjust(len(str(len(cmds))), '0')}】"
                    f"{f'{slot.namespace}::' if show_namespace else ''}{slot.header_display} : "
                    f"{slot.meta.description}"
                )
                for index, slot in enumerate(cmds)
                if query in str(slot.command)
            )
            if not command_string:
                return "查询失败！"
            return f"{command_string}\n{footer}"
        return slot.get_help()

    if not config.page_size:
        header = lang.require("manager", "help_header")
        command_string = "\n".join(
            (
                f" 【{str(index).rjust(len(str(len(cmds))), '0')}】"
                f"{f'{slot.namespace}::' if show_namespace else ''}{slot.header_display} : "
                f"{slot.meta.description}"
            )
            for index, slot in enumerate(cmds)
        )
        return f"{header}\n{command_string}\n{footer}"

    max_page = len(cmds) // config.page_size + 1
    if page < 1 or page > max_page:
        page = 1
    max_length = config.page_size
    if interactive:
        footer += "\n" + "输入 '<', 'a' 或 '>', 'd' 来翻页"

    def _(_page: int):
        header = (
            lang.require("manager", "help_header")
            + "\t"
            + lang.require("manager", "help_pages").format(current=_page, total=max_page)
        )
        command_string = "\n".join(
            (
                f" 【{str(index).rjust(len(str(_page * max_length)), '0')}】"
                f"{f'{slot.namespace}::' if show_namespace else ''}{slot.header_display} : "
                f"{slot.meta.description}"
            )
            for index, slot in enumerate(
                cmds[(_page - 1) * max_length : _page * max_length], start=(_page - 1) * max_length
            )
        )
        return f"{header}\n{command_string}\n{footer}"

    if not interactive:
        return _(page)

    def generator(_page: int):
        while True:
            resp = yield _(_page)
            if resp == "a" or resp == "<":
                _page -= 1
                if _page < 1:
                    _page = max_page
            elif resp == "d" or resp == ">":
                _page += 1
                if _page > max_page:
                    _page = 1
            else:
                return

    return generator(page)


disp = command.mount(help_cmd)


@disp.on_execute()
async def help_exec(arp: Arparma):
    return help_cmd_handle(arp)


@disp.handle()
async def help_handle(arp: Arparma, session: Session):
    resp = help_cmd_handle(arp, True)
    if isinstance(resp, str):
        return await session.send(resp)
    msg = await session.prompt(next(resp), timeout=15)
    while msg:
        try:
            msg = await session.prompt(resp.send(msg.extract_plain_text().strip().lower()), timeout=15)
        except StopIteration:
            return



from __future__ import annotations

import inspect
from os import PathLike
from pathlib import Path
from typing import Any, Callable, overload

from arclet.letoderea import es
from tarina import init_spec

from ..config import C, EntariConfig, config_model_validate
from ..event.plugin import PluginLoadedFailed, PluginLoadedSuccess, PluginUnloaded
from ..logger import log
from .model import PluginMetadata as PluginMetadata
from .model import RegisterNotInPluginError
from .model import RootlessPlugin as RootlessPlugin
from .model import StaticPluginDispatchError, _current_plugin
from .model import TE, TS, Plugin
from .model import keeping as keeping
from .module import import_plugin
from .module import package as package
from .module import requires as requires
from .service import plugin_service


def get_plugin(depth: int | None = None) -> Plugin:
    if plugin := _current_plugin.get(None):
        return plugin
    if depth is None:
        raise LookupError("no plugin context found")
    current_frame = inspect.currentframe()
    if current_frame is None:
        raise ValueError("Depth out of range")
    frame = current_frame
    d = depth + 1
    while d > 0:
        frame = frame.f_back
        if frame is None:
            raise ValueError("Depth out of range")
        d -= 1
    if (mod := inspect.getmodule(frame)) and "__plugin__" in mod.__dict__:
        return mod.__plugin__
    raise LookupError("no plugin context found")


def dispatch(event: type[TE], name: str | None = None):
    return get_plugin().dispatch(event, name=name)


def load_plugin(
    path: str, config: dict | None = None, recursive_guard: set[str] | None = None, prelude: bool = False
) -> Plugin | None:
    """
    以导入路径方式加载模块

    Args:
        path (str): 模块路径
        config (dict): 模块配置
        recursive_guard (set[str]): 递归保护
        prelude (bool): 是否为前置插件
    """
    if config is None:
        config = EntariConfig.instance.plugin.get(path)
    conf = config or {}
    if "$static" in conf:
        del conf["$static"]
    conf = conf.copy()
    if prelude:
        conf["$static"] = True
    if recursive_guard is None:
        recursive_guard = set()
    path = path.replace("::", "arclet.entari.builtins.")
    while path in plugin_service._subplugined:
        path = plugin_service._subplugined[path]
    if path in plugin_service.plugins:
        return plugin_service.plugins[path]
    if path in plugin_service._apply:
        return plugin_service._apply[path](conf)
    try:
        mod = import_plugin(path, config=conf)
        if not mod:
            path1 = f"entari_plugin_{path}"
            mod = import_plugin(path1, config=conf)
        if not mod:
            log.plugin.error(f"cannot found plugin <blue>{path!r}</blue>")
            es.publish(PluginLoadedFailed(path))
            return
        log.plugin.success(f"loaded plugin <blue>{mod.__name__!r}</blue>")
        if mod.__name__ in plugin_service._unloaded:
            if mod.__name__ in plugin_service._referents and plugin_service._referents[mod.__name__]:
                referents = plugin_service._referents[mod.__name__].copy()
                plugin_service._referents[mod.__name__].clear()
                for referent in referents:
                    if referent in recursive_guard:
                        continue
                    if referent in plugin_service.plugins:
                        log.plugin.debug(f"reloading <y>{mod.__name__}</y>'s referent <y>{referent!r}</y>")
                        unload_plugin(referent)
                        if not load_plugin(referent):
                            plugin_service._referents[mod.__name__].add(referent)
                        else:
                            recursive_guard.add(referent)
            plugin_service._unloaded.discard(mod.__name__)
        es.publish(PluginLoadedSuccess(mod.__name__))
        return mod.__plugin__
    except (ImportError, RegisterNotInPluginError, StaticPluginDispatchError) as e:
        log.plugin.error(f"failed to load plugin <blue>{path!r}</blue>: {e.args[0]}")
        es.publish(PluginLoadedFailed(path, e))
    except Exception as e:
        log.plugin.exception(f"failed to load plugin <blue>{path!r}</blue> caused by {e!r}", exc_info=e)
        es.publish(PluginLoadedFailed(path, e))


def load_plugins(dir_: str | PathLike | Path):
    path = dir_ if isinstance(dir_, Path) else Path(dir_)
    if not path.is_dir():
        raise NotADirectoryError(f"{path} is not a directory")
    for p in path.iterdir():
        if p.suffix in (".py", "") and p.stem not in {"__init__", "__pycache__"}:
            load_plugin(".".join(p.parts[:-1:1]) + "." + p.stem)


@init_spec(PluginMetadata)
def metadata(data: PluginMetadata):
    get_plugin()._metadata = data  # type: ignore


@overload
def plugin_config() -> dict[str, Any]: ...


@overload
def plugin_config(model_type: type[C]) -> C: ...


def plugin_config(model_type: type[C] | None = None):
    """获取当前插件的配置"""
    plugin = get_plugin()
    if model_type:
        return config_model_validate(model_type, plugin.config)
    return plugin.config


get_config = plugin_config


def declare_static():
    """声明当前插件为静态插件"""
    plugin = get_plugin()
    plugin.is_static = True
    if plugin._scope.subscribers:
        raise StaticPluginDispatchError("static plugin cannot dispatch events")


def add_service(serv: TS | type[TS]) -> TS:
    return get_plugin().service(serv)


def collect_disposes(*disposes: Callable[[], None]):
    return get_plugin().collect(*disposes)


def find_plugin(name: str) -> Plugin | None:
    if name in plugin_service.plugins:
        return plugin_service.plugins[name]
    if f"entari_plugin_{name}" in plugin_service.plugins:
        return plugin_service.plugins[f"entari_plugin_{name}"]


def find_plugin_by_file(file: str) -> Plugin | None:
    path = Path(file).resolve()
    for plugin in plugin_service.plugins.values():
        if plugin.module.__file__ == str(path):
            return plugin
        if plugin.module.__file__ and Path(plugin.module.__file__).parent == path:
            return plugin
        path1 = Path(path)
        while path1.parent != path1:
            if str(path1) == plugin.module.__file__:
                return plugin
            path1 = path1.parent
    return None


def unload_plugin(plugin: str):
    while plugin in plugin_service._subplugined:
        plugin = plugin_service._subplugined[plugin]
    if not (_plugin := find_plugin(plugin)):
        return False
    es.publish(PluginUnloaded(_plugin.id))
    _plugin.dispose()
    return True


listen = es.on



import asyncio
from asyncio.events import _get_running_loop  # type: ignore
from datetime import datetime, timedelta
from traceback import print_exc
from typing import Callable, Literal

from arclet.letoderea import es
from arclet.letoderea.typing import Contexts
from launart import Launart, Service, any_completed
from launart.status import Phase

from .plugin import RootlessPlugin, _current_plugin


class _ScheduleEvent:
    async def gather(self, context: Contexts):
        pass


pub = es.define(_ScheduleEvent, "entari.event/schedule")
scope = es.scope("entari.scheduler")
contexts: Contexts = {"$event": _ScheduleEvent()}  # type: ignore


class TimerTask:
    def __init__(self, supplier: Callable[[], timedelta], sub_id: str):
        self.sub_id = sub_id
        self.supplier = supplier
        self.handle = None

    def start(self, queue: asyncio.Queue):
        loop = asyncio.get_running_loop()
        self.handle = loop.call_later(self.supplier().total_seconds(), queue.put_nowait, self)

    def cancel(self):
        if self.handle and not self.handle.cancelled():
            self.handle.cancel()


class Scheduler(Service):

    def __init__(self):
        super().__init__()
        self.queue: asyncio.Queue[TimerTask] = asyncio.Queue()
        self.timers: dict[str, TimerTask] = {}

    @property
    def required(self) -> set[str]:
        return set()

    @property
    def stages(self) -> set[Phase]:
        return {"preparing", "blocking", "cleanup"}

    async def fetch(self):
        while True:
            timer = await self.queue.get()
            if timer.sub_id not in scope.subscribers:
                del self.timers[timer.sub_id]
                continue
            timer.start(self.queue)
            try:
                await scope.subscribers[timer.sub_id][0].handle(contexts.copy())
            except Exception:
                print_exc()

    def schedule(self, timer: Callable[[], timedelta], once: bool = False):

        def wrapper(func: Callable):
            sub = scope.register(func, temporary=once, publisher=pub)
            if plugin := _current_plugin.get():
                plugin.collect(sub.dispose)
            self.timers[sub.id] = TimerTask(timer, sub.id)
            if _get_running_loop():
                self.timers[sub.id].start(self.queue)
            return sub

        return wrapper

    async def launch(self, manager: Launart):
        async with self.stage("preparing"):
            for timer in self.timers.values():
                timer.start(self.queue)

        async with self.stage("blocking"):
            sigexit_task = asyncio.create_task(manager.status.wait_for_sigexit())
            fetch_task = asyncio.create_task(self.fetch())
            done, pending = await any_completed(sigexit_task, fetch_task)
            if sigexit_task in done:
                fetch_task.cancel()

        async with self.stage("cleanup"):
            for timer in self.timers.values():
                timer.cancel()

    id = "entari.scheduler"


scheduler = service = Scheduler()


@RootlessPlugin.apply("scheduler")
def _(plg: RootlessPlugin):
    plg.service(service)


def every_second():
    """每秒执行一次"""
    return lambda: timedelta(seconds=1)


def every_seconds(seconds: int):
    """每 seconds 秒执行一次

    Args:
        seconds (int): 距离下一次执行的时间间隔, 单位为秒
    """
    if seconds > 59 or seconds < 1:
        raise ValueError
    return lambda: timedelta(seconds=seconds)


def every_minute():
    """每分钟执行一次"""
    return lambda: timedelta(minutes=1)


def every_minutes(minutes: int):
    """每 minutes 分钟执行一次

    Args:
        minutes (int): 距离下一次执行的时间间隔, 单位为分
    """
    if minutes > 59 or minutes < 1:
        raise ValueError
    return lambda: timedelta(minutes=minutes)


def every_hour():
    """每小时执行一次"""
    return lambda: timedelta(hours=1)


def every_hours(hours: int):
    """每 hours 小时执行一次

    Args:
        hours (int): 距离下一次执行的时间间隔, 单位为小时
    """
    if hours > 23 or hours < 1:
        raise ValueError
    return lambda: timedelta(hours=hours)


def every_week():
    """每隔一周执行一次"""
    return lambda: timedelta(weeks=1)


def every_weeks(weeks: int):
    """每 weeks 周执行一次

    Args:
        weeks (int): 距离下一次执行的时间间隔, 单位为周
    """
    if weeks > 52 or weeks < 1:
        raise ValueError
    return lambda: timedelta(weeks=weeks)


def every_day():
    """每隔一天执行一次"""
    return lambda: timedelta(days=1)


def every_days(days: int):
    """每 days 天执行一次

    Args:
        days (int): 距离下一次执行的时间间隔, 单位为天
    """
    if days > 31 or days < 1:
        raise ValueError
    return lambda: timedelta(days=days)


def crontab(cron_str: str):
    """根据 cron 表达式执行

    Args:
        cron_str (str): cron 表达式
    """
    try:
        from croniter import croniter
    except ModuleNotFoundError:
        raise ImportError("Please install `croniter` first. Install with `pip install arclet-entari[cron]`") from None

    it = croniter(cron_str, datetime.now())

    return lambda iter=it: iter.get_next(datetime) - datetime.now()


def cron(pattern: str):
    return service.schedule(crontab(pattern))


def every(
    value: int = 1,
    mode: Literal["second", "minute", "hour"] = "second",
):
    _TIMER_MAPPING = {
        "second": every_seconds,
        "minute": every_minutes,
        "hour": every_hours,
    }
    return service.schedule(_TIMER_MAPPING[mode](value))


def invoke(delay: float):
    """延迟执行"""
    return service.schedule(lambda: timedelta(seconds=delay), once=True)



from __future__ import annotations

from collections.abc import Sequence
from dataclasses import dataclass, field
from pathlib import Path
import sys
from types import ModuleType
from typing import Any, Callable, TypeVar, overload
from weakref import finalize, proxy

from arclet.letoderea import Propagator, Provider, ProviderFactory, Scope, StepOut, Subscriber, es
from arclet.letoderea.publisher import Publisher, _publishers
from arclet.letoderea.typing import TTarget
from creart import it
from launart import Launart, Service
from tarina import ContextModel

from ..filter import parse
from ..logger import log
from .service import plugin_service

_current_plugin: ContextModel[Plugin] = ContextModel("_current_plugin")

T = TypeVar("T")
TE = TypeVar("TE")
TS = TypeVar("TS", bound=Service)
R = TypeVar("R")


class RegisterNotInPluginError(Exception):
    pass


class StaticPluginDispatchError(Exception):
    pass


class PluginDispatcher:
    def __init__(
        self,
        plugin: Plugin,
        event: type[TE],
        name: str | None = None,
    ):
        self.publisher = es.define(event, name)
        self.plugin = plugin
        self._event = event
        self.providers: list[Provider[Any] | ProviderFactory] = []
        self.propagators: list[Propagator] = []

    def waiter(
        self,
        *events: Any,
        providers: Sequence[Provider | type[Provider]] | None = None,
        priority: int = 15,
        block: bool = False,
    ) -> Callable[[TTarget[R]], StepOut[R]]:
        def wrapper(func: TTarget[R]):
            nonlocal events
            if not events:
                events = (self._event,)
            return StepOut(list(events), func, providers, priority, block)  # type: ignore

        return wrapper

    @overload
    def register(
        self,
        func: Callable[..., Any],
        *,
        priority: int = 16,
        providers: (
            Sequence[Provider[Any] | type[Provider[Any]] | ProviderFactory | type[ProviderFactory]] | None
        ) = None,
        temporary: bool = False,
    ) -> Subscriber: ...

    @overload
    def register(
        self,
        *,
        priority: int = 16,
        providers: (
            Sequence[Provider[Any] | type[Provider[Any]] | ProviderFactory | type[ProviderFactory]] | None
        ) = None,
        temporary: bool = False,
    ) -> Callable[[Callable[..., Any]], Subscriber]: ...

    def register(
        self,
        func: Callable[..., Any] | None = None,
        *,
        priority: int = 16,
        providers: (
            Sequence[Provider[Any] | type[Provider[Any]] | ProviderFactory | type[ProviderFactory]] | None
        ) = None,
        temporary: bool = False,
    ):
        _providers = providers or []
        wrapper = self.plugin._scope.register(
            priority=priority,
            providers=[*self.providers, *_providers],
            temporary=temporary,
            publisher=self.publisher,
        )

        def decorator(func1, /):
            self.plugin.validate(func1)
            sub = wrapper(func1)
            sub.propagates(*self.propagators)
            return sub

        if func:
            return decorator(func)
        return decorator

    @overload
    def once(
        self,
        func: Callable[..., Any],
        *,
        priority: int = 16,
        providers: (
            Sequence[Provider[Any] | type[Provider[Any]] | ProviderFactory | type[ProviderFactory]] | None
        ) = None,
    ) -> Subscriber: ...

    @overload
    def once(
        self,
        *,
        priority: int = 16,
        providers: (
            Sequence[Provider[Any] | type[Provider[Any]] | ProviderFactory | type[ProviderFactory]] | None
        ) = None,
    ) -> Callable[[Callable[..., Any]], Subscriber]: ...

    def once(
        self,
        func: Callable[..., Any] | None = None,
        *,
        priority: int = 16,
        providers: (
            Sequence[Provider[Any] | type[Provider[Any]] | ProviderFactory | type[ProviderFactory]] | None
        ) = None,
    ):
        if func:
            return self.register(func, priority=priority, providers=providers, temporary=True)
        return self.register(priority=priority, providers=providers, temporary=True)

    on = register
    handle = register

    def __call__(self, func):
        return self.register()(func)


@dataclass
class PluginMetadata:
    name: str
    author: list[str] = field(default_factory=list)
    version: str | None = None
    license: str | None = None
    urls: dict[str, str] | None = None
    description: str | None = None
    icon: str | None = None
    classifier: list[str] = field(default_factory=list)
    requirements: list[str] = field(default_factory=list)
    dependencies: list[str] = field(default_factory=list)
    config: Any | None = None

    # standards: list[str] = field(default_factory=list)
    # frameworks: list[str] = field(default_factory=list)
    # component_endpoints: list[str] = field(default_factory=list)


@dataclass
class Plugin:
    id: str
    module: ModuleType

    subplugins: set[str] = field(default_factory=set)
    config: dict[str, Any] = field(default_factory=dict)
    is_static: bool = False
    _metadata: PluginMetadata | None = None
    _is_disposed: bool = False
    _services: dict[str, Service] = field(init=False, default_factory=dict)
    _dispose_callbacks: list[Callable[[], None]] = field(init=False, default_factory=list)
    _scope: Scope = field(init=False)
    _extra: dict[str, Any] = field(default_factory=dict, init=False)  # extra metadata for inspection

    @property
    def available(self) -> bool:
        return not self._is_disposed

    @staticmethod
    def current() -> Plugin:
        try:
            return _current_plugin.get()  # type: ignore
        except LookupError:
            raise LookupError("no plugin context found") from None

    @property
    def metadata(self) -> PluginMetadata | None:
        return self._metadata

    def inject(self, *requires: str):
        plugin = self
        while plugin.id in plugin_service._subplugined:
            plugin = plugin_service.plugins[plugin_service._subplugined[plugin.id]]
        if plugin._metadata:
            plugin._metadata.requirements.extend(requires)
        return self

    def collect(self, *disposes: Callable[[], None]):
        self._dispose_callbacks.extend(disposes)
        return self

    def __post_init__(self):
        self._scope = es.scope(self.id)
        plugin_service.plugins[self.id] = self
        allow = self.config.pop("$allow", {})
        deny = self.config.pop("$deny", {})
        pat = {}
        if allow:
            pat["$and"] = allow
        if deny:
            pat["$not"] = deny
        if pat:
            self._scope.propagators.append(parse(pat))
        if "$static" in self.config:
            self.is_static = True
            self.config.pop("$static")
        if self.id not in plugin_service._keep_values:
            plugin_service._keep_values[self.id] = {}
        if self.id not in plugin_service._referents:
            plugin_service._referents[self.id] = set()
        finalize(self, self.dispose)

    def dispose(self):
        plugin_service._unloaded.add(self.id)
        if self._is_disposed:
            return
        self._is_disposed = True
        for serv in self._services.values():
            try:
                it(Launart).remove_component(serv)
            except ValueError:
                pass
        self._services.clear()
        if self.module.__spec__ and self.module.__spec__.cached:
            Path(self.module.__spec__.cached).unlink(missing_ok=True)
        sys.modules.pop(self.module.__name__, None)
        for callback in self._dispose_callbacks:
            callback()
        self._dispose_callbacks.clear()
        delattr(self.module, "__plugin__")
        if self.subplugins:
            subplugs = [i.removeprefix(self.id)[1:] for i in self.subplugins]
            subplugs = (subplugs[:3] + ["..."]) if len(subplugs) > 3 else subplugs
            log.plugin.debug(f"disposing sub-plugin <r>{', '.join(subplugs)}</r> of <y>{self.id}</y>")
            for subplug in self.subplugins:
                if subplug not in plugin_service.plugins:
                    continue
                try:
                    plugin_service.plugins[subplug].dispose()
                except Exception as e:
                    log.plugin.error(f"failed to dispose sub-plugin <r>{subplug}</r> caused by {e!r}")
                    plugin_service.plugins.pop(subplug, None)
            self.subplugins.clear()
        self._scope.dispose()
        self._scope.propagators.clear()
        del plugin_service.plugins[self.id]
        del self.module

    def dispatch(self, event: type[TE], name: str | None = None):
        if self.is_static:
            raise StaticPluginDispatchError("static plugin cannot dispatch events")
        return PluginDispatcher(self, event, name=name)

    @overload
    def use(
        self,
        pub: Any,
        *,
        priority: int = 16,
        providers: (
            Sequence[Provider[Any] | type[Provider[Any]] | ProviderFactory | type[ProviderFactory]] | None
        ) = None,
    ) -> Callable[[Callable[..., Any]], Subscriber]: ...

    @overload
    def use(
        self,
        pub: Any,
        func: Callable[..., Any],
        *,
        priority: int = 16,
        providers: (
            Sequence[Provider[Any] | type[Provider[Any]] | ProviderFactory | type[ProviderFactory]] | None
        ) = None,
    ) -> Subscriber: ...

    def use(
        self,
        pub: Any,
        func: Callable[..., Any] | None = None,
        *,
        priority: int = 16,
        providers: (
            Sequence[Provider[Any] | type[Provider[Any]] | ProviderFactory | type[ProviderFactory]] | None
        ) = None,
    ):
        if self.is_static:
            raise StaticPluginDispatchError("static plugin cannot use events by `Plugin.use`")
        if isinstance(pub, str):
            pid = pub.replace("::", "entari.event/")
        elif isinstance(pub, Publisher):
            pid = pub.id
        else:
            pid = getattr(pub, "__publisher__")
        if pid not in _publishers:
            raise LookupError(f"no publisher found: {pid}")
        disp = PluginDispatcher(self, _publishers[pid].target)
        if func:
            return disp.register(func=func, priority=priority, providers=providers)
        return disp.register(priority=priority, providers=providers)

    def validate(self, func):
        if func.__module__ != self.module.__name__:
            if "__plugin__" in func.__globals__ and func.__globals__["__plugin__"] is self:
                return
            raise RegisterNotInPluginError(
                f"Handler `{func.__qualname__}` should define "
                f"in the same module as the plugin: {self.module.__name__}. "
                f"Please use the `load_plugin({func.__module__!r})` or `requires({func.__module__!r})`"
                f"or `package({func.__module__!r})` before import it."
            )

    def proxy(self):
        return proxy(self.module)

    def subproxy(self, sub_id: str):
        return proxy(plugin_service.plugins[sub_id].module)

    def service(self, serv: TS | type[TS]) -> TS:
        if isinstance(serv, type):
            serv = serv()
        self._services[serv.id] = serv
        if plugin_service.status.blocking:
            it(Launart).add_component(serv)
        return serv


class RootlessPlugin(Plugin):

    @classmethod
    @overload
    def apply(cls, id: str) -> Callable[[Callable[[RootlessPlugin], Any]], Callable[[], None]]: ...

    @classmethod
    @overload
    def apply(cls, id: str, func: Callable[[RootlessPlugin], Any]) -> Callable[[], None]: ...

    @classmethod
    def apply(cls, id: str, func: Callable[[RootlessPlugin], Any] | None = None) -> Any:
        if not id.startswith("."):
            id = f".{id}"

        def dispose():
            if id in plugin_service.plugins:
                plugin_service.plugins[id].dispose()
            else:
                plugin_service._apply.pop(id, None)

        def wrapper(func: Callable[[RootlessPlugin], Any]):
            plugin_service._apply[id] = lambda config: cls(id, func, config)
            return dispose

        if func:
            return wrapper(func)
        return wrapper

    def __init__(self, id: str, func: Callable[[RootlessPlugin], Any], config: dict):
        super().__init__(id, ModuleType(id), config=config)
        setattr(self.module, "__plugin__", self)
        setattr(self.module, "__file__", func.__code__.co_filename)
        self.func = func
        token = _current_plugin.set(self)
        try:
            func(self)
        finally:
            _current_plugin.reset(token)

    def validate(self, func):
        pass


class KeepingVariable:
    def __init__(self, obj: T, dispose: Callable[[T], None] | None = None):
        self.obj = obj
        self._dispose = dispose
        try:
            setattr(self.obj, "__keeping__", True)
        except AttributeError:
            pass

    def dispose(self):
        if hasattr(self.obj, "dispose"):
            self.obj.dispose()  # type: ignore
        elif self._dispose:
            self._dispose(self.obj)
        del self.obj


def keeping(id_: str, obj: T, dispose: Callable[[T], None] | None = None) -> T:
    if not (plug := _current_plugin.get(None)):
        raise LookupError("no plugin context found")
    if id_ not in plugin_service._keep_values[plug.id]:
        plugin_service._keep_values[plug.id][id_] = KeepingVariable(obj, dispose)
    else:
        obj = plugin_service._keep_values[plug.id][id_].obj  # type: ignore
    return obj



from arclet.alconna import Arparma

from arclet.entari import MessageChain, command, metadata
from arclet.entari.command import Match

metadata(
    "echo",
    author=["RF-Tar-Railt <rf_tar_railt@qq.com>"],
    description="Echo the content",
)


@(
    command.command("echo <...content>", "显示消息")
    .option("escape", "-e|--escape # 发送转义消息")
    .option("unescape", "-E|--unescape # 发送反转义消息")
    .config(compact=True)
)
def echo(content: Match[MessageChain], arp: Arparma):
    if arp.find("unescape"):
        return MessageChain.of(content.result.extract_plain_text())
    elif arp.find("escape"):
        return str(content.result)
    else:
        return content.result



from dataclasses import dataclass
from typing import TYPE_CHECKING, Union

from arclet.letoderea import Contexts, es, provide
from satori.client import Account
from satori.model import MessageReceipt

from ..message import MessageChain

if TYPE_CHECKING:
    from ..session import Session


@dataclass
class SendRequest:
    account: Account
    channel: str
    message: MessageChain
    session: Union["Session", None] = None

    async def gather(self, context: Contexts):
        context["account"] = self.account
        context["channel"] = self.channel
        context["message"] = self.message
        if self.session:
            context["session"] = self.session

    __publisher__ = "entari.event/before_send"
    __result_type__: "type[bool | MessageChain]" = Union[bool, MessageChain]


before_send_pub = es.define(SendRequest)
before_send_pub.bind(provide(MessageChain, target="message"))


@dataclass
class SendResponse:
    account: Account
    channel: str
    message: MessageChain
    result: list[MessageReceipt]
    session: Union["Session", None] = None

    async def gather(self, context: Contexts):
        context["account"] = self.account
        context["channel"] = self.channel
        context["message"] = self.message
        context["result"] = self.result
        if self.session:
            context["session"] = self.session

    __publisher__ = "entari.event/send"


send_pub = es.define(SendResponse)
send_pub.bind(provide(MessageChain, target="message"))
send_pub.bind(provide(list, target="result"))
